/**
 * Copyright (C) 2003-2004 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#include <stdio.h>

#include "cedb.h"
#include "error.h"
#include "debug.h"

static BOOL firstSearched = FALSE;


/*
 * Opens the database specified by the given name, creating a new
 * database if 'create' is TRUE.
 *
 * @param volumeId the CEGUID of the volume the database has to be opened/created
 * @param dbName the database name
 * @param sort the property id to use as sort order
 * @param create if TRUE, if the database does not exist, it is create; if FALSE
 *               and the database does not exist, an invalid handle is retured
 * @param oid (OUT) database CEOID 
 */
HANDLE openDB (PCEGUID  volumeId,
               WCHAR*   dbName  ,
               CEPROPID sort    ,
               PCEOID   oid     )
{
    HANDLE db;                  // Open handle to the address database

    *oid = 0;
    db = CeOpenDatabaseEx (
            volumeId,            // Pointer to the mounted volume
            oid,                 // Location for the database identifier
            dbName,              // Database name
            sort,                // Sort order; 0 indicates to ignore
            CEDB_AUTOINCREMENT,  // Automatically increase seek pointer
            NULL);

#ifdef DEBUG
    if (db == INVALID_HANDLE_VALUE) {
        WCHAR msg[256];
        wsprintf(msg, TEXT("Error opening the database %s"), dbName);

        error(msg);
    } else {
        firstSearched = FALSE;
        wprintf(TEXT("Database %s opened; oid: %d, handle: %d\n"), dbName, *oid, db);
        
    }
#endif

    // Return the database handle.
    return db;
}

/*
 * Mounts the specified volume. If create is true, the volume is (re)created.
 * 
 * @param volume the volume pathname
 * @param create if TRUE the volume is recreate if it already exists or created
 *               if it does not exists
 * @param volumeId (OUT) the volum CEGUID
 *
 * Return value: TRUE if the volume has been created/mounted, FALSE otherwise
 */
BOOL mountDBVolume(WCHAR *volume, PCEGUID volumeId) {
    BOOL success = FALSE;

    success = CeMountDBVol(volumeId, volume, OPEN_ALWAYS);

    if (success == FALSE) {
        WCHAR msg[512];

        wsprintf(msg, TEXT("Unable to open volume %s"), volume);
        error(msg);
    }

    return success;
}


/*
 * Creates a new database in the given mouted volume.
 *
 * @param volumeId the volume where the database is created
 * @param dbName the database name
 * @param key the property id of the key field
 * @param status the property id of the status field
 */
CEOID createDB (PCEGUID  volumeId,
                WCHAR*   dbName  ,
                CEPROPID key     ,    
                CEPROPID status  ) {
    CEOID oid = 0;
    CEDBASEINFO info;  // Structure containing the 
                       // database data

    memset (&info, 0, sizeof(info));

    // Create the database with the following specified flags.
    // Create the database as uncompressed.
    // You can use CeSetDataBaseInfoEx to compress the database.
    info.dwFlags =
      CEDB_VALIDNAME        // szDbaseName is valid
      | CEDB_VALIDDBFLAGS   // HIWORD of dwFlag is valid
      | CEDB_VALIDSORTSPEC  // rgSortSpecs is valid
      | CEDB_NOCOMPRESS;    // The database is not compressed.

    // Assign the database name as MyDBase.
    wcscpy (info.szDbaseName, dbName);

    // Assign the database type.
    info.dwDbaseType = 0;
    info.wNumSortOrder = 2; 
    info.rgSortSpecs[0].dwFlags = 0;
    info.rgSortSpecs[0].propid = key;
    info.rgSortSpecs[1].dwFlags = 0;
    info.rgSortSpecs[1].propid = status;

    // Create database.
    oid = CeCreateDatabaseEx (volumeId, &info);

#ifdef DEBUG
    printf("Database created: %d\n", oid);
#endif

    return oid;
}

/*
 * Stores the given record into the database. If the record already exists,
 * it is updated, otherwise a new record is created.
 *
 * @param db the database handle
 * @param key the record key
 * @param nFields the number of fields to write
 * @param fields the fields to write
 *
 */
CEOID storeRecord(HANDLE db, CEPROPVAL* key, WORD nFields, CEPROPVAL *fields) {
    CEOID oid = FALSE;

    oid = findRecordByKey(db, key);

#ifdef DEBUG
    if (oid == 0) {
        wprintf(TEXT("Record %s not found\n"), key->val.lpwstr);
    } else {
        wprintf(TEXT("Record %s found: %d\n"), key->val.lpwstr, oid);
    }
#endif

    oid = CeWriteRecordProps(db, oid, nFields, fields);

    if (oid == 0) {
        error(TEXT("Error writing record"));
    } else {
#ifdef DEBUG
        printf("Record %d stored\n", oid);
#endif
    }

    return oid;
}

/*
 * Search a record with the given key. It returns the oid of the record found 
 * or 0 if no record with the given key is found.
 * 
 * @param key the record key 
 *
 */
CEOID findRecordByKey(HANDLE db, CEPROPVAL* key) {
    DWORD index;
    CEOID oid;

    CeSeekDatabase(db, CEDB_SEEK_BEGINNING, 0, &index);
    oid = CeSeekDatabase(db, CEDB_SEEK_VALUEFIRSTEQUAL, (DWORD)key, &index);

#ifdef DEBUG
    if (oid == 0) {
        error(TEXT("Record not found (may not be an error)"));
    }
#endif

    return oid;
}

/*
 * Search a record with the given status. It returns the oid of the record found 
 * or 0 if no record with the given status is found.
 * 
 * @param key the record key 
 *
 */
CEOID findRecordByStatus(HANDLE db, CEPROPVAL* state) {
    DWORD index;

    if (firstSearched == FALSE) {
#ifdef DEBUG
        wprintf(TEXT("First search of %s\n"), state->val.lpwstr);
#endif
        firstSearched = TRUE;
        return CeSeekDatabase(db, CEDB_SEEK_VALUEFIRSTEQUAL, (DWORD)state, &index);
    } else {
#ifdef DEBUG
        wprintf(TEXT("Searching next of %s\n"), state->val.lpwstr);
#endif
        return CeSeekDatabase(db, CEDB_SEEK_VALUENEXTEQUAL, (DWORD)state, &index);
    }
}
 


/*
 * Returns the number of record in the database specified by the
 * given oid.
 *
 * @param volumeId the volume containing the database
 * @param oid the database oid
 *
 */
DWORD getRecordCount(PCEGUID volumeId, CEOID oid) {
    CEOIDINFO info;

    if (CeOidGetInfoEx(volumeId, oid, &info) == FALSE) {
#ifdef DEBUG
       WCHAR msg[100];

       wsprintf(msg, TEXT("No info for oid %d"), oid);
       error(msg);

       return 0;
#endif
    }

    if (info.wObjType != OBJTYPE_DATABASE) {
#ifdef DEBUG
       WCHAR msg[100];

       wsprintf(msg, TEXT("The oid %d does not seem to be of type database (type: %d)"), oid, info.wObjType );
       error(msg);

       return 0;
#endif
    }

    return info.infDatabase.wNumRecords;
}

/*
 * Deletes the record specified by the given key.
 * 
 * @param key the record key
 *
 */
void deleteRecord(HANDLE db, CEPROPVAL* key) {
    CEOID oid = findRecordByKey(db, key);

#ifdef DEBUG
    printf("Deleting oid: %d\n", oid);
#endif

    if (oid != 0) {
        CeDeleteRecord(db, oid);
    }
}