/**
 * Copyright (C) 2003-2004 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */


package sync4j.syncclient.spds;

import java.net.*;
import java.io.*;
import java.util.Properties;

import sync4j.syncclient.spds.SyncMLClient;
import sync4j.syncclient.spds.util.StaticDataHelper;


/**
 *
 * The charSet is specified with the system property
 * <code>spds.charset</code>;
 * if <code>DEFAULT</code> value, the default charSet value is taken.<br>
 * if no value, the default charSet API <code>UTF-8</vode> value is taken.
 *
 * @author Stefano Fornari
 *
 * @version $Id: SyncMLClientImpl.java,v 1.1.1.1 2004/04/09 11:23:04 fabius Exp $
 *
 */
public final class SyncMLClientImpl implements SyncMLClient {

    // --------------------------------------------------------------- Constants

    private static final String API_CHARSET           = "UTF-8"         ;
    private static final String PROP_CHARSET          = "spds.charset"  ;
    private static final String KEY_DEFAULT_CHARSET   = "DEFAULT"       ;

    // --------------------------------------------------------------- Private data

    private final URL requestURL;

    // -------------------------------------------------------------- Properties

    /**
     * The charSet (encoding)
     */
    private String charSet = null;

    /**
     * Getter for property charSet.
     * @return Value of property charSet.
     */
    public String getCharSet() {
        return charSet;
    }

    /**
     * Setter for property charSet
     * @param charSet New value of property charSet.
    */
    public void setCharSet(String charSet) {
        this.charSet = charSet;
    }

    /**
     * The timeout (in milliseconds) for the request (it defaults to 1 minute)
     */
    private int timeout = 60000;

    /**
     * Getter for property timeout.
     * @return Value of property timeout.
     */
    public int getTimeout() {
        return timeout;
    }

    /**
     * Setter for property timeout. Default: 60 sec. 0 means forever
     * @param timeout New value of property timeout.
    */
    public void setTimeout(int timeout) {
        this.timeout = timeout;
    }

    /**
     * The proxy server. If null no proxy are used.
     */
    private String proxyHost = null;

    /** Getter for property proxyHost.
     * @return Value of property proxyHost.
     */
    public java.lang.String getProxyHost() {
        return proxyHost;
    }

    /** Setter for property proxyHost.
     * @param proxyHost New value of property proxyHost.
     */
    public void setProxyHost(java.lang.String proxyHost) {
        this.proxyHost = proxyHost;
    }

    /**
     * The proxy server port.
     * It defaults to 8080.
     */
    private int proxyPort = 8080;

    /** Getter for property proxyPort.
     * @return Value of property proxyPort.
     */
    public int getProxyPort() {
        return proxyPort;
    }

    /** Setter for property proxyPort.
     * @param proxyHost New value of property proxyPort.
     */
    public void setProxyPort(int proxyPort) {
        this.proxyPort = proxyPort;
    }

    /**
     * The proxy enable / disable setting
     * It defaults to false.
     */
    private boolean useProxy = false;

    /** Getter for property useProxy.
     * @return Value of property useProxy.
     */
    public boolean getUseProxy() {
        return useProxy;
    }

    /** Setter for property userProxy.
     * @param userProxy New value of property userProxy.
     */
    public void setUseProxy(boolean useProxy) {
        this.useProxy = useProxy;
    }


    // ------------------------------------------------------------ Constructors
    /**
     *
     * @param requestURL must be non-null
     *
     */
    public SyncMLClientImpl(final String requestURL) throws IOException
    {
        if (requestURL == null)
        {
            throw new NullPointerException("requestURL parameter is null");
        }

        this.requestURL = new URL(requestURL);

        setCharSet();

    }

    /**
     *
     * @param requestURL - NOT NULL
     *
     */
    public SyncMLClientImpl(final URL requestURL) throws IOException
    {
        if (requestURL == null)
        {
            throw new NullPointerException("requestURL parameter is null");
        }

        this.requestURL = requestURL;

        setCharSet();

    }

    // ---------------------------------------------------------- Public methods

    public String sendMessage(String request)
    throws IOException {

        Properties prop = System.getProperties();

        try {

            if (this.useProxy) {
                prop.put("http.proxyHost",proxyHost);
                prop.put("http.proxyPort",String.valueOf(proxyPort));
            }

            URLConnection conn = requestURL.openConnection();
            conn.setDoInput(true);
            conn.setDoOutput(true);
            conn.setRequestProperty ("Content-Type", "application/vnd.syncml+xml");
            conn.setRequestProperty ("Content-Length", String.valueOf(request.length()));

            StaticDataHelper.log("Connected to " + requestURL);

            OutputStream os = conn.getOutputStream();

            os.write(request.getBytes()); os.flush();

            StaticDataHelper.log("Message sent");
            StaticDataHelper.log("Reading response...");

            InputStream is = conn.getInputStream();

            int    contentLength = conn.getContentLength();
            String contentType   = conn.getContentType()  ;

            StaticDataHelper.log("contentLength: " + contentLength);
            StaticDataHelper.log("contentType: " + contentType);

            if (contentType == null)
            {
                throw new IOException("Content type is null");
            }

            byte[] buf = new byte[contentLength];


            int c = 0, b=0;

            while ((c<buf.length) && (b = is.read(buf, c, buf.length-c)) >= 0) {
                c+=b;
            }

            StaticDataHelper.log("Read: " + c);

            if (contentType.equals("application/vnd.syncml+xml") == false)
            {
                StaticDataHelper.log("response: " + new String(buf));
                throw new IOException( "Content type: "
                                     + contentType
                                     + " (should be "
                                     + "application/vnd.syncml+xml"
                                     + ")"
                                     );
            }

            //
            // Shut down the connection
            //
            os.close();
            is.close();

            conn = null;

            StaticDataHelper.log("Response read");

            if (!KEY_DEFAULT_CHARSET.equals(charSet)) {
                return new String(buf, charSet);
            } else {
                return new String(buf);
            }

        } catch (IOException e) {
            e.printStackTrace();
            throw e;
        }

    }

    public String toString()
    {
        return requestURL.toString();
    }

    /**
     * set charSet
     */
    private void setCharSet() {
        this.charSet = System.getProperty(PROP_CHARSET, API_CHARSET);
    }

}